r"""
Compute checksums.

(c) Tom de Geus, 2021, MIT
"""
import os

import numpy as np
import tqdm

from .relpath import add_prefix
from .yaml import read


def sha256(filename, size=2**10):
    r"""
    Get sha256 of a file.
    """

    import hashlib

    h = hashlib.sha256()

    with open(filename, "rb") as f:
        for byte_block in iter(lambda: f.read(size * h.block_size), b""):
            h.update(byte_block)
        return h.hexdigest()


def get(filepaths, yaml_hostinfo=None, hybrid=False, progress=False):
    r"""
    Compute the checksums of a list of files.

    :param list filepaths: List of file-paths.

    :param set yaml_hostinfo:
        File-path of a host-info file (see :py:mod:`shelephant.cli.hostinfo`).
        If specified the checksums are **not** computed, but exclusively read from the
        host-file. The user is responsible for keeping them up-to-date.

    :param bool hybrid:
        If ``True``, the function first tries to read from ``yaml_hostinfo``, and then
        computes missing items on the fly.

    :param bool progress: Show a progress-bar.

    :return:
        List of checksums, of same length as ``filepaths``.
        The entry is ``None`` if no checksum was found/read.
    """

    if type(filepaths) == str:
        filepaths = [filepaths]

    n = len(filepaths)
    ret = [None for i in range(n)]

    # Compute

    if not yaml_hostinfo:

        for i in tqdm.trange(n, disable=not progress, desc="Processing"):
            if os.path.isfile(filepaths[i]):
                ret[i] = sha256(filepaths[i])

        return ret

    # Read pre-computed

    data = read(yaml_hostinfo)
    files = data["files"]
    prefix = data["prefix"]
    check_sums = data["checksum"]
    check_paths = add_prefix(prefix, files)

    sorter = np.argsort(filepaths)
    source_paths = np.array(filepaths)[sorter]

    i = np.argsort(check_paths)
    check_paths = np.array(check_paths)[i]
    check_sums = np.array(check_sums)[i]

    test = np.in1d(source_paths, check_paths)
    idx = np.searchsorted(check_paths, source_paths)
    idx = np.where(test, idx, 0)
    ret = np.where(test, check_sums[idx], None)
    out = np.empty_like(ret)
    out[sorter] = ret
    ret = list(out)

    if hybrid:
        for i in tqdm.trange(n, disable=not progress, desc="Processing"):
            if ret[i] is None:
                if os.path.isfile(filepaths[i]):
                    ret[i] = sha256(filepaths[i])

    return ret



"""Copy files from remote, using earlier collected host-information.
Use ``shelephant_hostinfo`` to collect host-information.

By default, *rsync* is used to check if files are different.
*rsync* uses basic criteria such as file size and creation and modification date,
see `rsync manual <https://www.samba.org/ftp/rsync/rsync.html>`_.
This is fast but is only approximate.
In addition you can use:

*   ``--checksum``: Have *rsync* compare checksums of files.
*   ``--check-manual``: Do not use *rsync*.
    If ``--checksum`` is used, checksums are compared (that are optionally precomputed).

The default copy backend is ``rysnc -a --from-file="temp" host:source_dir dest_dir``.
Alternatively ``scp -p host:source_file dest_file`` can be used.
Typically, *rsync* will be faster, especially in copying a lot of small files.

:usage:

    shelephant_get [options]

    shelephant_get [options] <hostinfo.yaml>

:argument:

    YAML-file with host information. Default: shelephant_hostinfo.yaml

:options:

    -l, --local=arg
        Add local 'host' information to use precomputed checksums.

    --scp
        Use ``scp`` instead of ``rysnc`` as backend.

    -M, --check-manual
        Use internal algorithms instead of *rsync*.

    --colors=arg
        Select color scheme from: none, dark. [default: dark]

    -s, --summary
        Print summary (and no details unless specified).

    -d, --details
        Print details (and no summary unless specified).

    -f, --force
        Force overwrite of all existing (but not matching) files.

    --verbose
        Verbose all commands.

    -q, --quiet
        Do not print progress.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import argparse
import os
import shutil
import warnings

from .. import detail
from .. import rsync
from .. import scp
from .. import version
from .. import yaml
from .defaults import f_hostinfo


def main_impl():
    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-l", "--local")
    parser.add_argument("--scp", action="store_true")
    parser.add_argument("-M", "--check-manual", action="store_true")
    parser.add_argument("--colors", default="dark")
    parser.add_argument("-s", "--summary", action="store_true")
    parser.add_argument("-d", "--details", action="store_true")
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("-q", "--quiet", action="store_true")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("hostinfo", nargs="?", default=f_hostinfo)
    args = parser.parse_args()

    use_rsync = True
    if args.check_manual:
        use_rsync = False
    if not shutil.which("rsync"):
        warnings.warn("rsync not found, using internal fallback")
        use_rsync = False

    source = args.hostinfo
    data = yaml.read(source)

    if "host" not in data:

        detail.copy(
            copy_function=shutil.copy2,
            files=data["files"],
            src_dir=data["prefix"],
            dest_dir=os.path.dirname(source),
            checksum="checksum" in data,
            check_rsync=use_rsync,
            quiet=args.quiet,
            force=args.force,
            print_details=not (args.force or args.summary) or args.details,
            print_summary=not (args.force or args.details) or args.summary,
            print_all=args.details,
            theme_name=args.colors.lower(),
            yaml_hostinfo_src=source,
            yaml_hostinfo_dest=args.local,
        )

    elif args.scp:

        detail.copy_ssh(
            copy_function=scp.from_remote,
            use_rsync=False,
            host=data["host"],
            files=data["files"],
            src_dir=data["prefix"],
            dest_dir=os.path.dirname(source),
            checksum="checksum" in data,
            to_remote=False,
            check_rsync=use_rsync,
            quiet=args.quiet,
            force=args.force,
            print_details=not (args.force or args.summary) or args.details,
            print_summary=not (args.force or args.details) or args.summary,
            print_all=args.details,
            verbose=args.verbose,
            theme_name=args.colors.lower(),
            yaml_hostinfo_src=source,
            yaml_hostinfo_dest=args.local,
        )

    else:

        detail.copy_ssh(
            copy_function=rsync.from_remote,
            use_rsync=True,
            host=data["host"],
            files=data["files"],
            src_dir=data["prefix"],
            dest_dir=os.path.dirname(source),
            checksum="checksum" in data,
            to_remote=False,
            check_rsync=use_rsync,
            quiet=args.quiet,
            force=args.force,
            print_details=not (args.force or args.summary) or args.details,
            print_summary=not (args.force or args.details) or args.summary,
            print_all=args.details,
            verbose=args.verbose,
            theme_name=args.colors.lower(),
            yaml_hostinfo_src=source,
            yaml_hostinfo_dest=args.local,
        )


def main():

    try:
        main_impl()
    except Exception as e:
        print(e)
        return 1


if __name__ == "__main__":

    main()



r"""Collect file information from location (on a remote host).

:usage:

    shelephant_hostinfo [options]

:options:

    -o, --output=arg
        Output YAML-file. [default: shelephant_hostinfo.yaml].

    -r, --host=arg
        Host-name.

    -p, --prefix=arg
        Directory (on host) from which to copy.

    -f, --files=[arg]
        YAML-file with list of files (on remote). [default: shelephant_dump.yaml]

    -c, --checksum=[arg]
        YAML-file with checksums (on remote). [default: shelephant_checksum.yaml]

    --files-key=arg
        Path in the YAML-file, separated by "/". [default: /]

    --checksum-key=arg
        Path in the YAML-file, separated by "/". [default: /]

    --ignore
        Skip basic checks.

    -i, --input=[arg]
        Use host and prefix from existing file. [default: shelephant_hostinfo.yaml]

    --remove=arg...
        Remove paths (and checksums) from existing file.

    --force
        Overwrite output file without prompt.

    --verbose
        Verbose all commands.

    -h, --help
        Show help.

    --version
        Show version.


(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import argparse
import os
import shutil
import tempfile

import numpy as np

from .. import scp
from .. import version
from .. import yaml
from .defaults import f_checksum
from .defaults import f_dump
from .defaults import f_hostinfo


def remove(data, rm):

    files = np.array(data["files"])
    sorter = np.argsort(files)
    files = files[sorter]
    rm = np.sort(rm)

    if not np.all(np.in1d(rm, files)):
        raise OSError("One or more remove paths not found")

    idx = np.searchsorted(files, rm)
    keep = np.setdiff1d(np.arange(len(files)), sorter[idx])

    files = data["files"]
    checksum = data["checksum"]
    data["files"] = [files[i] for i in keep]
    data["checksum"] = [checksum[i] for i in keep]

    return data


def main_impl():

    # Parse command-line arguments

    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-o", "--output", default=f_hostinfo)
    parser.add_argument("-r", "--host", default=None)
    parser.add_argument("-p", "--prefix", default=None)
    parser.add_argument("-f", "--files", default=None, nargs="?", const=f_dump)
    parser.add_argument("-c", "--checksum", default=None, nargs="?", const=f_checksum)
    parser.add_argument("--files-key", default="/")
    parser.add_argument("--checksum-key", default="/")
    parser.add_argument("--ignore", action="store_true")
    parser.add_argument("-i", "--input", default=None, nargs="?", const=f_hostinfo)
    parser.add_argument("--remove", nargs="*")
    parser.add_argument("--force", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("-v", "--version", action="version", version=version)
    args = parser.parse_args()

    # Separate mode: remove paths and quit

    if args.remove:
        filename = args.output
        if args.input:
            filename = args.input
        data = yaml.read(filename)
        data = remove(data, args.remove)
        yaml.dump(args.output, data, args.force)
        return 0

    # Extract basic information from command-line arguments

    data = {}

    if args.host:
        data["host"] = args.host

    if args.prefix:
        data["prefix"] = args.prefix

    if args.input:
        overwrite = yaml.read(args.input)
        for item in ["host", "prefix"]:
            if item in overwrite:
                data[item] = overwrite[item]

    # Basic IO-checks

    if not args.ignore:
        if "host" in data and "prefix" not in data:
            raise OSError("Specify hostname and prefix")

    if "host" not in data and "prefix" not in data:
        for item in [args.files, args.checksum]:
            if item:
                data["prefix"] = os.path.dirname(item)
                break

    # Create temporary file to download to

    if "host" in data:
        temp_dir = tempfile.mkdtemp()
        temp_file = os.path.join(temp_dir, f_hostinfo)

    # Read files and checksums

    args_dict = vars(args)

    for item in ["files", "checksum"]:

        if args_dict[item]:

            key = list(filter(None, args_dict[item + "_key"].split("/")))
            filename = args_dict[item]

            if "host" in data:
                scp.from_remote(
                    data["host"],
                    os.path.join(data["prefix"], filename),
                    temp_file,
                    args.verbose,
                )
                filename = temp_file

            data[item] = yaml.read_item(filename, key)

    # Run basic checks

    if not args.ignore:
        for item in ["prefix"]:
            if item not in data:
                raise OSError(f"Please specify {item:s}")
        if "checksum" in data:
            if len(data["files"]) != len(data["checksum"]):
                raise OSError("Number of checksums does not match number of files")

    # Clean-up: remove temporary file

    if "host" in data:
        shutil.rmtree(temp_dir)

    # Write output

    if "files" not in data:
        data["files"] = []

    yaml.dump(args.output, data, args.force)


def main():

    try:
        main_impl()
    except Exception as e:
        print(e)
        return 1


if __name__ == "__main__":

    main()



"""ssh_test
    Run SSH test.

Usage:
    ssh_test [options] --host=N --prefix=N

Options:
    -r, --host=N        Host name.
    -p, --prefix=M      Path on host.
        --version       Print version.
    -g, --help          Print this help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import os
import subprocess

import docopt

import shelephant


def run(cmd):
    print(cmd)
    return subprocess.check_output(cmd, shell=True).decode("utf-8")


args = docopt.docopt(__doc__, version=shelephant.version)

# shelephant_send - local checksum

operations = [
    "bar.txt -> bar.txt",
    "foo.txt == foo.txt",
]

output = run(
    (
        "shelephant_hostinfo -o myssh_send/shelephant_hostinfo.yaml --force "
        '--host "{:s}" --prefix "{:s}" -f -c'
    ).format(args["--host"], os.path.join(args["--prefix"], "myssh_get"))
)

output = run(
    "shelephant_send --detail --colors none --force "
    "myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml"
)

output = list(filter(None, output.split("\n")))
assert output == operations


# shelephant_send - basic

operations = [
    "bar.txt == bar.txt",
    "foo.txt -> foo.txt",
]

output = run(
    (
        "shelephant_hostinfo -o myssh_send/shelephant_hostinfo.yaml --force "
        '--host "{:s}" --prefix "{:s}" -f'
    ).format(args["--host"], os.path.join(args["--prefix"], "myssh_get"))
)

output = run(
    "shelephant_send --detail --colors none --force "
    "myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml"
)

output = list(filter(None, output.split("\n")))
assert output == operations

# shelephant_diff

operations = {
    "<-": [],
    "->": [],
    "!=": [],
    "==": ["bar.txt", "foo.txt"],
}

output = run(
    "shelephant_diff -f --yaml shelephant_diff.yaml "
    "myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml"
)

output = shelephant.yaml.read("shelephant_diff.yaml")

for key in operations:
    assert output[key] == operations[key]

# shelephant_get - basic

operations = [
    "bar.txt -> bar.txt",
    "foo.txt == foo.txt",
]

output = run(
    (
        "shelephant_hostinfo -o myssh_get/shelephant_hostinfo.yaml --force "
        '--host "{:s}" --prefix "{:s}" -f -c'
    ).format(args["--host"], os.path.join(args["--prefix"], "myssh_send"))
)

output = run("shelephant_get --detail --colors none --force myssh_get/shelephant_hostinfo.yaml")

output = list(filter(None, output.split("\n")))
assert output == operations

os.remove("myssh_get/bar.txt")



import os
import shutil
import subprocess
import unittest

import numpy as np

import shelephant


def run(cmd, verbose=False):
    return subprocess.check_output(cmd, shell=True).decode("utf-8")


class Test_tools(unittest.TestCase):
    def test_flatten(self):

        arg = [1, [2, 2, 2], 4]
        ret = [1, 2, 2, 2, 4]

        self.assertEqual(ret, shelephant.convert.flatten(arg))

    def test_squash(self):

        arg = {"foo": [1, 2], "bar": {"foo": [3, 4], "bar": 5}}
        ret = [1, 2, 3, 4, 5]

        self.assertEqual(ret, shelephant.convert.squash(arg))

    def test_deepest_dirs(self):

        dirnames = [
            "foo/bar",
            "foo/bar/mydir",
            "bar/mydir",
            "bar",
            "bar/mydir2",
            "foo/bar/mydir/deep",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow",
        ]

        ret = shelephant.path.filter_deepest(dirnames)

        d = [
            "bar/mydir",
            "bar/mydir2",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow",
        ]

        self.assertEqual(sorted(ret), sorted(d))


class Test_checksum(unittest.TestCase):
    def test_basic(self):

        with open("foo.txt", "w") as file:
            file.write("foo")

        with open("bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -f foo.txt bar.txt")
        run("shelephant_checksum -f -q shelephant_dump.yaml")
        data = shelephant.yaml.read_item("shelephant_checksum.yaml")

        keys = [
            "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae",
            "fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9",
        ]

        self.assertEqual(data, keys)

        os.remove("foo.txt")
        os.remove("bar.txt")
        os.remove("shelephant_dump.yaml")
        os.remove("shelephant_checksum.yaml")

    def test_hybrid(self):

        with open("foo.txt", "w") as file:
            file.write("foo")

        with open("bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -f foo.txt")
        run("shelephant_checksum -f -q")
        run("shelephant_hostinfo --force -f -c")

        run("shelephant_dump -f foo.txt bar.txt")
        run("shelephant_checksum -f -q -l shelephant_hostinfo.yaml")
        data = shelephant.yaml.read_item("shelephant_checksum.yaml")

        keys = [
            "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae",
            "fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9",
        ]

        self.assertEqual(data, keys)

        os.remove("foo.txt")
        os.remove("bar.txt")
        os.remove("shelephant_dump.yaml")
        os.remove("shelephant_checksum.yaml")
        os.remove("shelephant_hostinfo.yaml")

    def test_recursive(self):

        letters = ["a", "b", "c", "d", "e", "f", "g"]

        for letter in letters:
            with open(f"{letter:s}.txt", "w") as file:
                file.write(letter)

        files = [f"{letter:s}.txt" for letter in letters]

        keys = [
            "ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb",  # a
            "3e23e8160039594a33894f6564e1b1348bbd7a0088d42c4acb73eeaed59c009d",  # b
            "2e7d2c03a9507ae265ecf5b5356885a53393a2029d241394997265a1a25aefc6",  # c
            "18ac3e7343f016890c510e93f935261169d9e3f565436429830faf0934f4f8e4",  # d
            "3f79bb7b435b05321651daefd374cdc681dc06faa65e374e38337b88ca046dea",  # e
            "252f10c83610ebca1a059c0bae8255eba2f95be4d1d7bcfa89d7248a82d9f111",  # f
            "cd0aa9856147b6c5b4ff2b7dfee5da20aa38253099ef1b4a64aced233c9afe29",  # g
        ]

        run("shelephant_dump -f -s {:s}".format(" ".join(files)))
        run("shelephant_checksum -f -q")

        data = shelephant.yaml.read_item("shelephant_checksum.yaml")

        self.assertEqual(data, keys)

        ifiles = np.arange(len(files))
        np.random.shuffle(ifiles)

        run("shelephant_dump -f {:s}".format(" ".join([files[i] for i in ifiles[:3]])))
        run("shelephant_checksum -f -q")
        run("shelephant_hostinfo --force -f -c")

        np.random.shuffle(ifiles)

        run("shelephant_dump -f {:s}".format(" ".join([files[i] for i in ifiles])))
        run("shelephant_checksum -f -q -l shelephant_hostinfo.yaml")
        run("shelephant_hostinfo --force -f -c")

        run("shelephant_dump -f -s {:s}".format(" ".join(files)))
        run("shelephant_checksum -f -q -l shelephant_hostinfo.yaml")

        data = shelephant.yaml.read_item("shelephant_checksum.yaml")

        self.assertEqual(data, keys)

        for file in files:
            os.remove(file)

        os.remove("shelephant_dump.yaml")
        os.remove("shelephant_checksum.yaml")
        os.remove("shelephant_hostinfo.yaml")


class Test_dump(unittest.TestCase):
    def test_basic(self):

        with open("myfile_foo.txt", "w") as file:
            file.write("foo")

        with open("myfile_bar.txt", "w") as file:
            file.write("bar")

        for dirname in ["mydir"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mydir")

        with open("mydir/myfile_foo.txt", "w") as file:
            file.write("foo")

        with open("mydir/myfile_bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -f -s -o dump_1.yaml myfile_foo.txt myfile_bar.txt")
        run("shelephant_dump -f -s -o dump_2.yaml *.txt")
        run("shelephant_dump -f -s -o dump_3.yaml -c \"find . -maxdepth 1 -iname 'myfile_*.txt'\"")
        run("shelephant_dump -f -s -o mydir/dump_4.yaml mydir/*.txt")

        with open("dump_1.yaml") as file:
            dump_1 = file.read()

        with open("dump_2.yaml") as file:
            dump_2 = file.read()

        with open("dump_3.yaml") as file:
            dump_3 = file.read()

        with open("mydir/dump_4.yaml") as file:
            dump_4 = file.read()

        self.assertEqual(dump_1, dump_2)
        self.assertEqual(dump_1, dump_3)
        self.assertEqual(dump_1, dump_4)

        os.remove("myfile_foo.txt")
        os.remove("myfile_bar.txt")
        os.remove("dump_1.yaml")
        os.remove("dump_2.yaml")
        os.remove("dump_3.yaml")
        shutil.rmtree("mydir")

    def test_append(self):

        with open("foo.txt", "w") as file:
            file.write("foo")

        with open("bar.txt", "w") as file:
            file.write("bar")

        with open("foo.pdf", "w") as file:
            file.write("foo")

        with open("bar.pdf", "w") as file:
            file.write("bar")

        run("shelephant_dump -f foo.txt bar.txt")
        run("shelephant_dump -a foo.pdf bar.pdf")

        self.assertEqual(
            shelephant.yaml.read("shelephant_dump.yaml"),
            ["foo.txt", "bar.txt", "foo.pdf", "bar.pdf"],
        )

        os.remove("foo.txt")
        os.remove("bar.txt")
        os.remove("foo.pdf")
        os.remove("bar.pdf")
        os.remove("shelephant_dump.yaml")


class Test_extract(unittest.TestCase):
    def test_single_path(self):

        data = {
            "foo": ["foo.txt", "bar.txt"],
            "bar": ["foo.pdf", "bar.pdf"],
            "key": ["foo.key", "bar.key"],
        }

        shelephant.yaml.dump("dump.yaml", data, force=True)

        run('shelephant_extract -f dump.yaml "foo"')

        self.assertEqual(shelephant.yaml.read("dump.yaml"), ["foo.txt", "bar.txt"])

        os.remove("dump.yaml")

    def test_multiple_paths(self):

        data = {
            "foo": ["foo.txt", "bar.txt"],
            "bar": ["foo.pdf", "bar.pdf"],
            "key": ["foo.key", "bar.key"],
            "sub": {
                "foo": ["foo.txt", "bar.txt"],
                "bar": ["foo.pdf", "bar.pdf"],
                "key": ["foo.key", "bar.key"],
            },
        }

        shelephant.yaml.dump("dump.yaml", data, force=True)

        run('shelephant_extract -f dump.yaml "/sub/foo" "foo"')

        self.assertEqual(
            shelephant.yaml.read("dump.yaml"),
            {"foo": ["foo.txt", "bar.txt"], "sub": {"foo": ["foo.txt", "bar.txt"]}},
        )

        os.remove("dump.yaml")

    def test_multiple_paths_squash(self):

        data = {
            "foo": ["foo.txt", "bar.txt"],
            "bar": ["foo.pdf", "bar.pdf"],
            "key": ["foo.key", "bar.key"],
            "sub": {
                "foo": ["foo2.txt", "bar2.txt"],
                "bar": ["foo2.pdf", "bar2.pdf"],
                "key": ["foo2.key", "bar2.key"],
            },
        }

        shelephant.yaml.dump("dump.yaml", data, force=True)

        run('shelephant_extract -f --squash dump.yaml "/sub/foo" "foo"')

        self.assertEqual(
            shelephant.yaml.read("dump.yaml"),
            ["foo2.txt", "bar2.txt", "foo.txt", "bar.txt"],
        )

        os.remove("dump.yaml")


class Test_merge(unittest.TestCase):
    def test_basic(self):

        with open("foo.txt", "w") as file:
            file.write("foo")

        with open("bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -o main.yaml foo.txt")
        run("shelephant_dump -o branch.yaml bar.txt")
        run("shelephant_merge -f branch.yaml main.yaml")

        self.assertEqual(shelephant.yaml.read("main.yaml"), ["foo.txt", "bar.txt"])

        os.remove("foo.txt")
        os.remove("bar.txt")
        os.remove("main.yaml")
        os.remove("branch.yaml")

        os.mkdir("dira")
        os.mkdir("dirb")

        with open("dira/foo.txt", "w") as file:
            file.write("foo")

        with open("dira/bar.txt", "w") as file:
            file.write("bar")

        with open("dirb/foo.txt", "w") as file:
            file.write("foo")

        with open("dirb/bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -o dira/dump.yaml dira/foo.txt dira/bar.txt")
        run("shelephant_dump -o dirb/dump.yaml dirb/foo.txt dirb/bar.txt")
        run("shelephant_merge -f dira/dump.yaml dirb/dump.yaml")

        self.assertEqual(
            shelephant.yaml.read("dirb/dump.yaml"),
            ["foo.txt", "bar.txt", "../dira/foo.txt", "../dira/bar.txt"],
        )

        shutil.rmtree("dira")
        shutil.rmtree("dirb")


class Test_hostinfo(unittest.TestCase):
    def test_basic(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -f -s -o mysrc/files.yaml mysrc/*.txt")
        run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        run(
            "shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml"
        )
        run("shelephant_get -f -q mydest/hostinfo.yaml")
        run("shelephant_dump -f -s -o mydest/files.yaml mydest/*.txt")
        run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")

    def test_remove(self):

        with open("foo.txt", "w") as file:
            file.write("foo")

        with open("bar.txt", "w") as file:
            file.write("bar")

        keys = [
            "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae",
            "fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9",
        ]

        run("shelephant_dump -f foo.txt bar.txt")
        run("shelephant_checksum -f -q")
        run("shelephant_hostinfo --force -f -c")
        run("shelephant_hostinfo --force --remove bar.txt")
        data = shelephant.yaml.read("shelephant_hostinfo.yaml")

        self.assertEqual(data["files"], ["foo.txt"])
        self.assertEqual(data["checksum"], [keys[0]])

        os.remove("foo.txt")
        os.remove("bar.txt")
        os.remove("shelephant_dump.yaml")
        os.remove("shelephant_checksum.yaml")
        os.remove("shelephant_hostinfo.yaml")


class Test_get(unittest.TestCase):
    def test_basic(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt -> foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        output = run(
            "shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml"
        )
        output = run("shelephant_get -f -d -q --colors none mydest/hostinfo.yaml")

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")

    def test_partial(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        shutil.copy("mysrc/foo.txt", "mydest/foo.txt")

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt == foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        output = run(
            "shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml"
        )
        output = run("shelephant_get -f -d -q --colors none mydest/hostinfo.yaml")

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")

    def test_partial_localchecksum(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        with open("mysrc/car.txt", "w") as file:
            file.write("car")

        with open("mysrc/dog.txt", "w") as file:
            file.write("dog")

        shutil.copy("mysrc/foo.txt", "mydest/foo.txt")
        shutil.copy("mysrc/dog.txt", "mydest/dog.txt")

        operations = [
            "bar.txt -> bar.txt",
            "car.txt -> car.txt",
            "dog.txt == dog.txt",
            "foo.txt == foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        output = run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")
        output = run(
            "shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml"
        )
        output = run(
            "shelephant_hostinfo -o mydest/local.yaml -f mydest/files.yaml -c mydest/checksum.yaml"
        )
        output = run(
            "shelephant_get -f -d -q --colors none -l mydest/local.yaml mydest/hostinfo.yaml"
        )

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump -f -s -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")

    def test_partial_rsync(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        shutil.copy2("mysrc/foo.txt", "mydest/foo.txt")

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt == foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml")
        output = run("shelephant_get -f -d -q --colors none mydest/hostinfo.yaml")

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")


class Test_send(unittest.TestCase):
    def test_basic(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        with open("mydest/foobar.txt", "w") as file:
            file.write("foobar")

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt -> foo.txt",
        ]

        run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")
        run("shelephant_hostinfo --force -f mydest/files.yaml -c mydest/checksum.yaml")
        output = run(
            "shelephant_send -f -d -q --colors none mysrc/files.yaml shelephant_hostinfo.yaml"
        )

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        os.remove("mydest/foobar.txt")

        run("shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt")
        run("shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")
        os.remove("shelephant_hostinfo.yaml")

    def test_empty_remote(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt -> foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_hostinfo --force -o hostinfo.yaml -p mydest")
        output = run("shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml")

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml")
        output = run("shelephant_checksum -f -q -o mysrc/checksum.yaml mysrc/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")
        os.remove("hostinfo.yaml")

    def test_partial(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        shutil.copy("mysrc/foo.txt", "mydest/foo.txt")

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt == foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        output = run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")
        output = run("shelephant_hostinfo --force -f mydest/files.yaml -c mydest/checksum.yaml")
        output = run(
            "shelephant_send -f -d -q --colors none mysrc/files.yaml shelephant_hostinfo.yaml"
        )

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")
        os.remove("shelephant_hostinfo.yaml")

    def test_partial_localchecksum(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        with open("mysrc/car.txt", "w") as file:
            file.write("car")

        with open("mysrc/dog.txt", "w") as file:
            file.write("dog")

        shutil.copy("mysrc/foo.txt", "mydest/foo.txt")
        shutil.copy("mysrc/dog.txt", "mydest/dog.txt")

        operations = [
            "bar.txt -> bar.txt",
            "car.txt -> car.txt",
            "dog.txt == dog.txt",
            "foo.txt == foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        output = run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")
        output = run("shelephant_hostinfo --force -f mydest/files.yaml -c mydest/checksum.yaml")
        output = run(
            "shelephant_hostinfo --force -o local.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml"
        )
        output = run(
            " ".join(
                [
                    "shelephant_send",
                    "-f",
                    "-d",
                    "-q",
                    "--colors none",
                    "-l local.yaml",
                    "mysrc/files.yaml",
                    "shelephant_hostinfo.yaml",
                ]
            )
        )

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")
        os.remove("shelephant_hostinfo.yaml")
        os.remove("local.yaml")

    def test_partial_rsync(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        shutil.copy2("mysrc/foo.txt", "mydest/foo.txt")

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt == foo.txt",
        ]

        output = run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        output = run("shelephant_hostinfo --force -o hostinfo.yaml -p mydest")
        output = run("shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml")

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt")
        output = run("shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")
        os.remove("hostinfo.yaml")


class Test_mv(unittest.TestCase):
    def test_basic(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -o mysrc/files.yaml mysrc/*.txt")
        run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        run("shelephant_mv -f -q mysrc/files.yaml mydest")
        run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")


class Test_cp(unittest.TestCase):
    def test_basic(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -o mysrc/files.yaml mysrc/*.txt")
        run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")
        run("shelephant_cp -f -q mysrc/files.yaml mydest")
        run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )
        self.assertTrue(os.path.isfile("mysrc/foo.txt"))
        self.assertTrue(os.path.isfile("mysrc/bar.txt"))

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")

    def test_rsync(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.log", "w") as file:
            file.write("foo")

        with open("mysrc/bar.log", "w") as file:
            file.write("bar")

        shutil.copy2("mysrc/foo.log", "mydest/foo.log")

        operations = [
            "bar.log -> bar.log",
            "foo.log == foo.log",
        ]

        output = run("shelephant_dump -o mysrc/files.yaml mysrc/*.log")
        output = run("shelephant_cp -f -d -q --colors none mysrc/files.yaml mydest")

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        output = run("shelephant_dump --sort -o mydest/files.yaml mydest/*.log")
        output = run("shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml")
        output = run("shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml")

        self.assertEqual(
            shelephant.yaml.read("mysrc/files.yaml"),
            shelephant.yaml.read("mydest/files.yaml"),
        )
        self.assertEqual(
            shelephant.yaml.read("mysrc/checksum.yaml"),
            shelephant.yaml.read("mydest/checksum.yaml"),
        )
        self.assertTrue(os.path.isfile("mysrc/foo.log"))
        self.assertTrue(os.path.isfile("mysrc/bar.log"))

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")

    def test_nested(self):

        for dirname in ["mysrc", "mybak"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.makedirs("mysrc/foo/foo/foo")

        with open("mysrc/foo.log", "w") as file:
            file.write("foo")

        with open("mysrc/foo/foo.log", "w") as file:
            file.write("foo")

        with open("mysrc/foo/foo/foo.log", "w") as file:
            file.write("foo")

        with open("mysrc/foo/foo/foo/foo.log", "w") as file:
            file.write("foo")

        operations = [
            "mysrc/foo.log             -> mysrc/foo.log",
            "mysrc/foo/foo.log         -> mysrc/foo/foo.log",
            "mysrc/foo/foo/foo.log     -> mysrc/foo/foo/foo.log",
            "mysrc/foo/foo/foo/foo.log -> mysrc/foo/foo/foo/foo.log",
        ]

        output = run("shelephant_dump -f -s `find . -iname '*.log'`")
        output = run("shelephant_cp -f -d -q --colors none mybak")

        self.assertEqual(list(filter(None, output.split("\n"))), operations)

        self.assertTrue(os.path.isfile("mybak/mysrc/foo.log"))
        self.assertTrue(os.path.isfile("mybak/mysrc/foo/foo.log"))
        self.assertTrue(os.path.isfile("mybak/mysrc/foo/foo/foo.log"))
        self.assertTrue(os.path.isfile("mybak/mysrc/foo/foo/foo/foo.log"))

        shutil.rmtree("mysrc")
        shutil.rmtree("mybak")


class Test_rm(unittest.TestCase):
    def test_basic(self):

        with open("foo.txt", "w") as file:
            file.write("foo")

        with open("bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -f foo.txt bar.txt")
        run("shelephant_rm -f shelephant_dump.yaml")

        self.assertFalse(os.path.isfile("foo.txt"))
        self.assertFalse(os.path.isfile("bar.txt"))

        os.remove("shelephant_dump.yaml")


class Test_diff(unittest.TestCase):
    def test_basic(self):

        for dirname in ["mysrc", "mydest"]:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir("mysrc")
        os.mkdir("mydest")

        with open("mysrc/foo.txt", "w") as file:
            file.write("foo")

        with open("mysrc/bar.txt", "w") as file:
            file.write("bar")

        with open("mydest/bar.txt", "w") as file:
            file.write("foobar")

        with open("mydest/foobar.txt", "w") as file:
            file.write("foobar")

        run("shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt")
        run("shelephant_dump --sort -o mydest/files.yaml mydest/*.txt")
        run("shelephant_diff mysrc/files.yaml mydest/files.yaml --yaml mysrc/diff.yaml")

        data = shelephant.yaml.read("mysrc/diff.yaml")
        expect = {
            "==": [],
            "!=": ["bar.txt"],
            "->": ["foo.txt"],
            "<-": ["foobar.txt"],
        }

        self.assertDictEqual(data, expect)

        shutil.rmtree("mysrc")
        shutil.rmtree("mydest")


class Test_parse(unittest.TestCase):
    def test_basic(self):

        with open("foo.txt", "w") as file:
            file.write("foo")

        with open("bar.txt", "w") as file:
            file.write("bar")

        run("shelephant_dump -f foo.txt bar.txt")
        output = run("shelephant_parse shelephant_dump.yaml")

        self.assertEqual(list(filter(None, output.split("\n"))), ["- foo.txt", "- bar.txt"])

        os.remove("shelephant_dump.yaml")
        os.remove("foo.txt")
        os.remove("bar.txt")


if __name__ == "__main__":

    unittest.main()



r"""
Copy & query using *rsync*.

(c) Tom de Geus, 2021, MIT
"""
import os
import re
import subprocess
import tempfile

import numpy as np
import tqdm

from .external import exec_cmd


def _rsync(source_dir, dest_dir, files, verbose=False, progress=True):
    r"""
    Copy files to a destination using ``rsync -a --files-from``.

    :param str source_dir: Source directory.
    :param str dest_dir: Source directory.
    :param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param bool verbose: Verbose commands.
    :param bool progress: Show progress bar.
    """

    with tempfile.TemporaryDirectory() as temp_dir:

        temp_file = os.path.join(temp_dir, "rsync.txt")

        with open(temp_file, "w") as file:
            file.write("\n".join(files))

        # Run without printing output

        if not progress:

            cmd = 'rsync -a --files-from="{files:s}" "{source_dir:s}" "{dest_dir:s}"'.format(
                source_dir=source_dir, dest_dir=dest_dir, files=temp_file
            )

            return exec_cmd(cmd, verbose)

        # Run while printing output

        cmd = 'rsync -aP --files-from="{files:s}" "{source_dir:s}" "{dest_dir:s}"'.format(
            source_dir=source_dir, dest_dir=dest_dir, files=temp_file
        )

        if verbose:
            print(cmd)

        pbar = tqdm.tqdm(total=len(files))
        sbar = tqdm.tqdm(unit="B", unit_scale=True)

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)

        for line in iter(process.stdout.readline, b""):
            line = line.decode("utf-8")
            if re.match(r"(.*)(xf)([e]?)(r\#)([0-9])(.*)(to\-ch)([e]?[c]?)(k\=)([0-9])(.*)", line):
                e = int(list(filter(None, line.split(" ")))[-6].replace(",", ""))
                pbar.update()
                sbar.update(e)


def from_remote(
    hostname,
    source_dir,
    dest_dir,
    files,
    verbose=False,
    progress=True,
):
    r"""
    Copy files from a remote system. Uses: ``rsync -a --files-from``.

    :param str hostname: Hostname.
    :param str source_dir: Source directory.
    :param str dest_dir: Source directory.
    :param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param bool verbose: Verbose commands.
    :param bool progress: Show progress bar.
    """

    return _rsync(
        source_dir=hostname + ":" + source_dir,
        dest_dir=dest_dir,
        files=files,
        verbose=verbose,
        progress=progress,
    )


def to_remote(
    hostname,
    source_dir,
    dest_dir,
    files,
    verbose=False,
    progress=True,
):
    r"""
    Copy files to a remote system. Uses: ``rsync -a --files-from``.

    :param str hostname: Hostname.
    :param str source_dir: Source directory.
    :param str dest_dir: Source directory.
    :param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param bool verbose: Verbose commands.
    :param bool progress: Show progress bar.
    """

    return _rsync(
        source_dir=source_dir,
        dest_dir=hostname + ":" + dest_dir,
        files=files,
        verbose=verbose,
        progress=progress,
    )


def diff(
    source_dir: str,
    dest_dir: str,
    files: list[str],
    checksum: bool = False,
    verbose: bool = False,
) -> np.array:
    r"""
    Check if files are different using *rsync*.
    *rsync* uses basic criteria such as file size and creation and modification date.
    This is much faster than using checksums but is only approximate.
    See `rsync manual <https://www.samba.org/ftp/rsync/rsync.html>`_.

    :param str source_dir: Source directory.
    :param str dest_dir: Source directory.
    :param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param checksum: Use checksum to test file difference.
    :param verbose: Verbose commands.
    """

    with tempfile.TemporaryDirectory() as temp_dir:

        temp_file = os.path.join(temp_dir, "rsync.txt")
        files = [os.path.normpath(file) for file in files]

        with open(temp_file, "w") as file:
            file.write("\n".join(files))

        # Run without printing output

        opt = "-nai"

        if checksum:
            opt += "c"

        cmd = 'rsync {opt:s} --files-from="{files:s}" "{source_dir:s}" "{dest_dir:s}"'.format(
            source_dir=source_dir, dest_dir=dest_dir, files=temp_file, opt=opt
        )

        lines = list(filter(None, exec_cmd(cmd, verbose).split("\n")))
        lines = [line for line in lines if line[1] == "f"]

        if len(lines) == 0:
            return {
                "skip": np.ones((len(files)), dtype=np.bool),
                "create": np.zeros((len(files)), dtype=np.bool),
                "overwrite": np.zeros((len(files)), dtype=np.bool),
            }

        check_paths = [line.split(" ")[1] for line in lines]
        mode = np.zeros((len(check_paths)), dtype=np.int16)

        for i, line in enumerate(lines):
            # todo: split send `<` and recieve `>`?
            # ref: https://stackoverflow.com/a/12037164/2646505
            if line[0] == ">" or line[0] == "<":
                if line[2] == "+":
                    mode[i] = 1  # create
                else:
                    mode[i] = 2  # overwrite
            elif line[0] == ".":
                pass
            else:
                raise OSError(f'Unknown cryptic output "{line:s}"')

        sorter = np.argsort(files)
        source_paths = np.array(files, dtype=str)[sorter]

        i = np.argsort(check_paths)
        check_paths = np.array(check_paths, dtype=str)[i]
        mode = mode[i]

        test = np.in1d(source_paths, check_paths)

        idx = np.searchsorted(check_paths, source_paths)
        idx = np.where(test, idx, 0)
        ret = np.where(test, mode[idx], 0)
        ret = ret.astype(np.int16)
        out = np.empty_like(ret)
        out[sorter] = ret

        return {
            "skip": out == 0,
            "create": out == 1,
            "overwrite": out == 2,
        }



r"""
Query using ssh.

(c) Tom de Geus, 2021, MIT
"""
from .external import exec_cmd


def file_exists(host, source, verbose=False):
    r"""
    Check if a file exists on a remote system. Uses ``ssh``.
    """

    cmd = 'ssh {host:s} test -f "{source:s}" && echo found || echo not found'.format(
        host=host, source=source
    )

    ret = exec_cmd(cmd, verbose)

    if ret == "found":
        return True

    return False



def theme(theme=None):
    r"""
    Return dictionary of colors.

    .. code-block:: python

        {
            'new' : '...',
            'overwrite' : '...',
            'skip' : '...',
            'bright' : '...',
        }

    :param str theme: Select color-theme.

    :rtype: dict
    """

    if theme == "dark":
        return {
            "new": "1;32",
            "overwrite": "1;31",
            "skip": "1;30",
            "bright": "1;37",
        }

    return {
        "new": "",
        "overwrite": "",
        "skip": "",
        "bright": "",
    }


class String:
    r"""
    Rich string.

    .. note::

        All options are attributes, that can be modified at all times.

    .. note::

        Available methods:

        *   ``A.format()`` :  Formatted string.
        *   ``str(A)`` : Unformatted string.
        *   ``A.isnumeric()`` : Return if the "data" is numeric.
        *   ``int(A)`` : Dummy integer.
        *   ``float(A)`` : Dummy float.

    :type data: str, None
    :param data: The data.

    :type width: None, int
    :param width: Print width (formatted print only).

    :type color: None, str
    :param color: Print color, e.g. "1;32" for bold green (formatted print only).

    :type align: ``'<'``, ``'>'``
    :param align: Print alignment (formatted print only).

    :type dummy: 0, int, float
    :param dummy: Dummy numerical value.

    :methods:


    """

    def __init__(self, data, width=None, align="<", color=None, dummy=0):

        self.data = data
        self.width = width
        self.color = color
        self.align = align
        self.dummy = dummy

    def format(self):
        r"""
        Return formatted string: align/width/color are applied.
        """

        if self.width and self.color:
            fmt = "\x1b[{color:s}m{{0:{align:s}{width:d}.{width:d}s}}\x1b[0m".format(
                **self.__dict__
            )
        elif self.width:
            fmt = "{{0:{align:s}{width:d}.{width:d}s}}".format(**self.__dict__)
        elif self.color:
            fmt = "\x1b[{color:s}m{{0:{align:s}s}}\x1b[0m".format(**self.__dict__)
        else:
            fmt = "{{0:{align:s}s}}".format(**self.__dict__)

        return fmt.format(str(self))

    def isnumeric(self):
        r"""
        Return if the "data" is numeric : always zero for this class.
        """
        return False

    def __str__(self):
        return str(self.data)

    def __int__(self):
        return int(self.dummy)

    def __float__(self):
        return float(self.dummy)

    def __repr__(self):
        return str(self)

    def __lt__(self, other):
        return str(self) < str(other)



"""Dump filenames to a YAML-file.

:usage:

    shelephant_dump [options] <file>...

:arguments:

    Files to dump. By default the filenames are written relative to the output file.

:options:

    -o, --output=arg
        Output YAML-file. [default: shelephant_dump.yaml]

    -a, --append
        Append existing file.

    -c, --command
        Interpret the input as a command (instead of as filenames).

    --cwd=arg
        Directory to run the command in.

    --abspath
        Store absolute paths (default: relative to the output file).

    -s, --sort
        Sort filenames.

    -f, --force
        Overwrite output file without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import argparse
import os
import subprocess

from .. import version
from .. import yaml
from .defaults import f_dump


def main_impl():
    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-o", "--output", default=f_dump)
    parser.add_argument("-a", "--append", action="store_true")
    parser.add_argument("-c", "--command", action="store_true")
    parser.add_argument("--cwd", type=str)
    parser.add_argument("--abspath", action="store_true")
    parser.add_argument("-s", "--sort", action="store_true")
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("file", nargs="+")
    args = parser.parse_args()

    prefix = os.path.dirname(args.output)
    files = args.file

    if args.command:
        command = " ".join(files)
        files = (
            subprocess.check_output(command, shell=True, cwd=args.cwd).decode("utf-8").split("\n")
        )
        files = sorted(list(filter(None, files)))
        if args.cwd is not None:
            files = [os.path.join(args.cwd, file) for file in files]

    if args.abspath:
        files = [os.path.abspath(file) for file in files]
    else:
        files = [os.path.relpath(file, prefix) for file in files]

    if args.sort:
        files = sorted(files)

    if args.append:

        main = yaml.read(args.output)
        if type(main) != list:
            raise OSError('Can only append a "flat" file')

        files = main + files
        args.force = True

    yaml.dump(args.output, files, args.force)


def main():

    try:
        main_impl()
    except Exception as e:
        print(e)
        return 1


if __name__ == "__main__":

    main()



"""Extract a field from a YAML-file.

Unless you use --no-path, the function assumes that all data are paths,
and changes all relative paths from being relative to <input.yaml>
to being relative to --output.

:usage:

    shelephant_extract [options] <input.yaml> [<key>...]

:arguments:

    <input.yaml>
        The file to read.

    <key>
        The keys to read from the file.

:options:

    -o, --output=arg
        Output file. (default: <input.yaml>)

    --no-path
        Do not interpret data as paths.

    -s, --squash
        Squash fields into a single field.

    -f, --force
        Overwrite output file without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import argparse
import functools
import os

import mergedeep

from .. import convert
from .. import relpath
from .. import version
from .. import yaml


def main_impl():
    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-o", "--output")
    parser.add_argument("--no-path", action="store_true")
    parser.add_argument("-s", "--squash", action="store_true")
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("input")
    parser.add_argument("key", nargs="*")
    args = parser.parse_args()

    input_dir = os.path.dirname(args.input)
    output = args.output if args.output else args.input
    output_dir = os.path.dirname(output)
    data = {}

    if len(args.key) == 0:
        args.key = ["/"]

    for key in args.key:
        key = list(filter(None, key.split("/")))
        files = yaml.read_item(args.input, key)
        if not args.no_path:
            files = relpath.chroot(files, input_dir, output_dir)
        if len(args.key) == 1:
            yaml.dump(output, files, args.force)
            return 0
        container = functools.reduce(lambda x, y: {y: x}, key[:-1], {key[-1]: files})
        mergedeep.merge(data, container)

    if args.squash:
        data = convert.squash(data)

    yaml.dump(output, data, args.force)


def main():

    try:
        main_impl()
    except Exception as e:
        print(e)
        return 1


if __name__ == "__main__":

    main()
